class Base:
    def __init__(self, stage, username=None, key=None):
        self.__version__ = "0.5.6"  # update here and in setup.py
        if stage == "develop":
            self.headers = {"Authorization": username + "," + key, "Api_key": username + "," + key}
            self.url = "https://test-api.dcentralab.com/"  # DUB backend URL
            self.w3i_url = "https://test-api.web3index.com/"  # Web3Index backend URL
        if stage == "staging":
            self.headers = {"Authorization": username + "," + key, "Api_key": username + "," + key}
            self.url = "https://staging.dcentralab.com/"
            self.w3i_url = "https://staging-api.web3index.com/"
        if stage == "preprod":
            self.headers = {"Authorization": username + "," + key, "Api_key": username + "," + key}
            self.url = "https://preprod-api.dcentralab.com/"
            self.w3i_url = "https://preprod-api.web3index.com/"
        if stage == "main":
            self.headers = {"Authorization": username + "," + key, "Api_key": username + "," + key}
            self.url = "https://api.dcentralab.com/"
            self.w3i_url = "https://api.web3index.com/"
        if stage == "staging-api":
            self.headers = {"Authorization": username + "," + key, "Api_key": username + "," + key}
            self.url = "https://staging-api.dcentralab.com/"
            self.w3i_url = "https://staging-api.web3index.com/"


class DapiError(Exception):
    def __init__(self, message=None, status_code=None, response=None, exception=None):
        super().__init__(message)
        self.status_code = status_code
        self.response = response
        self.exception = exception



from typing import List, Dict, Optional, Union
from dcentrapi.requests_dappi import requests_post
from dcentrapi.Base import Base


class EventPoller(Base):
    def register_user(self, user_name: str, collection_name: str):
        url = self.url + "event_poller/user/register"
        data = {
            "user_name": user_name,
            "collection_name": collection_name,
        }
        response = requests_post(url=url, json=data, headers=self.headers)
        return response.json()

    def subscribe_contract(
        self,
        contract_address: str,
        contract_name: str,
        abi: List[dict],
        chain_id: str,
        deployment_block_number: int,
        subscribed_events: List[str],
        webhook_url: Optional[str] = None,
    ):
        url = self.url + "event_poller/contract/subscribe"
        data = {
            "contract_name": contract_name,
            "contract_address": contract_address,
            "abi": abi,
            "chain_id": chain_id,
            "deployment_block_number": deployment_block_number,
            "webhook_url": webhook_url,
            "subscribed_events": subscribed_events,
        }
        response = requests_post(url=url, json=data, headers=self.headers)
        return response.json()

    def update_contract(
        self,
        contract_address: str,
        chain_id: str,
        updated_name: Optional[str] = None,
        updated_abi: Optional[List[dict]] = None,
        updated_subscribed_events: Optional[List[str]] = None,
        updated_webhook_url: Optional[str] = None,
    ):
        url = self.url + "event_poller/contract/update"
        data = {
            "contract_address": contract_address,
            "chain_id": chain_id,
            "updated_name": updated_name,
            "updated_abi": updated_abi,
            "updated_subscribed_events": updated_subscribed_events,
            "updated_webhook_url": updated_webhook_url,
        }
        response = requests_post(url=url, json=data, headers=self.headers)
        return response.json()

    def get_contracts(
        self,
        chain_contract_map: Optional[Dict[str, Optional[Union[str, List[str]]]]] = None,
        page_size: Optional[int] = 500,
        page_number: Optional[int] = 1,
    ):
        url = self.url + "event_poller/contracts"
        data = {
            "chain_contract_map": chain_contract_map,
            "page_size": page_size,
            "page_number": page_number,
        }
        response = requests_post(url=url, json=data, headers=self.headers)
        return response.json()

    def deactivate_contracts(self, chain_contract_map: Dict[str, Optional[Union[str, List[str]]]]):
        url = self.url + "event_poller/contracts/deactivate"
        data = {
            "chain_contract_map": chain_contract_map,
        }
        response = requests_post(url=url, json=data, headers=self.headers)
        return response.json()

    def get_events(
        self,
        chain_contract_map: Optional[Dict[str, Optional[Union[str, List[str]]]]] = None,
        event_args: Optional[Dict[str, Optional[List[str]]]] = None,
        event_names: Optional[List[str]] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        user_web3_addresses: Optional[List[str]] = None,
        block_number: Optional[int] = None,
        page_size: Optional[int] = 500,
        page_number: Optional[int] = 1,
    ):
        url = self.url + "event_poller/events"
        data = {
            "chain_contract_map": chain_contract_map,
            "event_args": event_args,
            "event_names": event_names,
            "start_time": start_time,
            "end_time": end_time,
            "user_web3_addresses": user_web3_addresses,
            "block_number": block_number,
            "page_size": page_size,
            "page_number": page_number,
        }
        response = requests_post(url=url, json=data, headers=self.headers)
        return response.json()



import datetime
import logging
import traceback
from time import sleep
from dcentrapi.Base import Base, DapiError
from dcentrapi.requests_dappi import requests_get

logger = logging.getLogger()
logger.setLevel("INFO")


class HackMitigation(Base):
    def are_addresses_blacklisted(self, addresses: [str], check_malcon: bool = True, check_sanctioned: bool = False):
        url = self.url + "generic_freeze_signal/are_addresses_blacklisted"
        data = {
            "addresses": addresses,
            "check_malcon": check_malcon,
            "check_sanctioned": check_sanctioned,
        }
        response = None
        tries = 7
        while tries > 0:
            logger.info(f"{tries} more tries")
            before = datetime.datetime.now()
            logger.info(f"before request: {before}")
            try:
                response = requests_get(url, params=data, headers=self.headers)

                after = datetime.datetime.now()
                logger.info(f"after request: {after}")
                logger.info(f"time elapsed: {after-before}")

                if response.status_code == 200:
                    # Parse the JSON data using the json() method
                    response_data = response.json()
                    logger.info(f"data={response_data}")  # This will print the parsed JSON data as a Python dictionary
                    return response_data
                else:
                    msg = f"Request failed with status code: {response.status_code}"
                    if hasattr(response, "error_message"):
                        msg += f" and error message: {response.error_message}"
                    logger.error(msg)
                    tries -= 1
                    logger.info("sleeping for 1.75 secs")
                    sleep(1.75)

            except Exception as e:
                return DapiError(response=response.__dict__, exception=f"e: {e}, traceback: {traceback.format_exc()}")

        return DapiError(response=response.__dict__, exception="Stop retry after 7 tries")



# import requests
# from eth_abi import encode
# from web3 import Web3
# from dcentrapi.Base import Base
#
#
# class MerkleTreeNode:
#     def __init__(self, value):
#         self.left = None
#         self.right = None
#         self.value = value
#         self.hashValue = Web3.solidityKeccak(["bytes"], [value])
#
#
# def verify_leaf(root: str, leaf: tuple, proof: list, encoding: str) -> bool:
#     encoded_leaf = encode([encoding], [leaf])
#     computed_hash = MerkleTreeNode(value=encoded_leaf).hashValue
#     for proof_elem in proof:
#
#         if computed_hash <= proof_elem:
#             encoded = encode(["(bytes32,bytes32)"], [(computed_hash, proof_elem)])
#         else:
#             encoded = encode(["(bytes32,bytes32)"], [(proof_elem, computed_hash)])
#         computed_hash = MerkleTreeNode(value=encoded).hashValue
#
#     return computed_hash.hex() == root
#
#
# class MerkleTree(Base):
#
#     def __init__(self, stage, username, key):
#         super().__init__(stage, username, key)
#         url = "https://test-api.dcentralab.com/auth"
#
#         response = requests.get(url, headers=self.headers)
#         if response.status_code != 200:
#             raise "Unauthorized, MerkleTree instance cannot be used"
#
#     nodes = []
#     layers = []
#     leaf_index_map = {}
#
#     def build_tree(self, leaves, encoding):
#         if len(leaves) == 0:
#             raise "no leaves, aborting"
#
#         self.nodes = []
#         for i in range(len(leaves)):
#             leaf = leaves[i]
#             encoded_leaf = encode([encoding], [leaf])
#             self.nodes.append(MerkleTreeNode(encoded_leaf))
#             self.leaf_index_map[leaf] = i
#         self.layers = [self.nodes]
#
#         while len(self.nodes) != 1:
#             temp = []
#             for i in range(0, len(self.nodes), 2):
#                 node1 = self.nodes[i]
#                 if i + 1 < len(self.nodes):
#                     node2 = self.nodes[i + 1]
#                 else:
#                     temp.append(self.nodes[i])
#                     break
#                 if node1.hashValue <= node2.hashValue:
#                     encoded = encode(["(bytes32,bytes32)"], [(node1.hashValue, node2.hashValue)])
#                 else:
#                     encoded = encode(["(bytes32,bytes32)"], [(node2.hashValue, node1.hashValue)])
#
#                 parent = MerkleTreeNode(encoded)
#
#                 parent.left = node1
#                 parent.right = node2
#                 temp.append(parent)
#             self.nodes = temp
#             self.layers.append(temp)
#         return self.nodes[0], self.layers
#
#     def get_proof(self, leaf: tuple) -> [MerkleTreeNode]:
#         proof = []
#         index = self.leaf_index_map[leaf]
#         for layer in self.layers:
#             is_right_node = index % 2
#             pair_index = index - 1 if is_right_node else index + 1
#
#             if pair_index < len(layer):
#                 proof.append(layer[pair_index])
#
#             index = int(index / 2)
#
#         return proof
#
#



from typing import List
from dcentrapi.Base import Base
from dcentrapi.requests_dappi import requests_get


class TokenPrices(Base):

    def get_token_prices_by_network_name_and_address(self, network_names: List[str], token_addresses: List[str]):
        url = self.w3i_url + "tokenPricesByNetworkAndAddress"
        data = {
            "network_names": network_names,
            "token_addresses": token_addresses,
        }
        response = requests_get(url, params=data, headers=self.headers)
        return response.json()

    def get_token_prices_by_chain_id_and_address(self, chain_ids: List[str], token_addresses: List[str]):
        url = self.w3i_url + "tokenPricesByNetworkAndAddress"
        data = {
            "chain_ids": chain_ids,
            "token_addresses": token_addresses,
        }
        response = requests_get(url, params=data, headers=self.headers)
        return response.json()



# import traceback
# from dcentrapi.Base import Base, DapiError
# from dcentrapi.requests_dappi import requests_get
# from requests.exceptions import RequestException, HTTPError
# from json.decoder import JSONDecodeError
#
# # TODO: implmenent add_gas_fee_to_tx_payload which will enable dapi users to send a tx_payload + network chain id and
# #  have it filled with the proper gas fees
#
#
# class GasMonitor(Base):
#     def get_optimal_gas_price(self, network_name, minutes, stats=None, values=None):
#         url = self.url + "gas_monitor/optimal_gas_price_for_network"
#         data = {
#             "network_name": network_name,
#             "minutes": minutes,
#             "stats": stats,
#             "values": values,
#         }
#         response = None
#         try:
#             response = requests_get(url, params=data, headers=self.headers)
#             # Check response status code before attempting to parse JSON
#             response.raise_for_status()  # This will raise HTTPError for bad status codes
#             return response.json()  # Attempt to parse JSON
#         except JSONDecodeError as json_error:
#             # Handle JSON parsing errors specifically
#             error_message = f"Failed to parse JSON response: {json_error}"
#             raise DapiError(
#                 message=error_message,
#                 status_code=response.status_code if response is not None else None,
#                 response=response.text if response is not None else None,
#                 exception=json_error,
#             )
#         except HTTPError as http_error:
#             # Handle HTTP errors
#             error_message = f"HTTP error occurred: {http_error}"
#             raise DapiError(
#                 message=error_message,
#                 status_code=response.status_code if response is not None else None,
#                 response=response.text if response is not None else None,
#                 exception=http_error,
#             )
#         except RequestException as request_error:
#             # Handle other requests-related errors (e.g., network issues)
#             error_message = f"Request failed: {request_error}"
#             raise DapiError(
#                 message=error_message,
#                 status_code=response.status_code if response is not None else None,
#                 exception=request_error,
#             )
#         except Exception as e:
#             # Catch-all for any other exceptions
#             raise DapiError(
#                 message="An unexpected error occurred", exception=f"e: {e}, traceback: {traceback.format_exc()}"
#             )



from dcentrapi.Base import Base
from dcentrapi.Base import DapiError
from dcentrapi.eventPolling import EventPolling
from dcentrapi.rpcAggregation import RpcAggregation
from dcentrapi.tokenPrices import TokenPrices
from dcentrapi.txSimulation import TxSimulation
from dcentrapi.gas_fee_estimator import GasFeeEstimator
from dcentrapi.event_poller import EventPoller
from dcentrapi.hackMitigation import HackMitigation
from dcentrapi.requests_dappi import requests_get, requests_post

# from dcentrapi.gasMonitor import GasMonitor
# from dcentrapi.merkleTree import MerkleTree, MerkleTreeNode



from typing import List
from dcentrapi.Base import Base
from dcentrapi.requests_dappi import requests_post


# See: https://docs.tenderly.co/simulations-and-forks/intro-to-simulations
class TxSimulation(Base):

    # In functions below:
    # `sender` is the address that will sign the transaction(s)
    #     (This gets turned into a `from` attribute on the tx(s). Not using `from` directly because it is a Python reserved word.)
    # `tx` is a built transaction output from web3.py
    # `tx_bundle` is a list of built transactions output from web3.py

    # See: https://docs.tenderly.co/simulations-and-forks/simulation-api/using-simulation-api
    def simulate_transaction_single(
        self,
        sender: str,
        tx: dict,
    ):
        url = self.url + "simulateTransactions"
        data = {"sender": sender, "tx_single": tx}
        response = requests_post(url, json=data, headers=self.headers)
        return response.json()

    # Format of tx_bundle is [tx0, tx1, tx2...]
    # See: https://docs.tenderly.co/simulations-and-forks/simulation-api/simulation-bundles
    def simulate_transaction_bundle(
        self,
        sender: str,
        tx_bundle: List[dict],
    ):
        url = self.url + "simulateTransactions"
        data = {"sender": sender, "tx_bundle": tx_bundle}
        response = requests_post(url, json=data, headers=self.headers)
        return response.json()



import traceback
from dcentrapi.Base import Base, DapiError
from dcentrapi.requests_dappi import requests_get
from requests.exceptions import RequestException, HTTPError
from json.decoder import JSONDecodeError


class GasFeeEstimator(Base):
    def get_gas_fee_estimation(self, chain_id: str, strategy: str = "market"):
        url = self.url + "gas_fee_estimator/get_gas_fee_estimation"
        data = {
            "chain_id": chain_id,
            "strategy": strategy,
        }

        response = None
        try:
            response = requests_get(url, params=data, headers=self.headers)
            # Check response status code before attempting to parse JSON
            response.raise_for_status()  # This will raise HTTPError for bad status codes
            return response.json()  # Attempt to parse JSON
        except JSONDecodeError as json_error:
            # Handle JSON parsing errors specifically
            error_message = f"Failed to parse JSON response: {json_error}"
            raise DapiError(
                message=error_message,
                status_code=response.status_code if response is not None else None,
                response=response.text if response is not None else None,
                exception=json_error,
            )
        except HTTPError as http_error:
            # Handle HTTP errors
            error_message = f"HTTP error occurred: {http_error}"
            raise DapiError(
                message=error_message,
                status_code=response.status_code if response is not None else None,
                response=response.text if response is not None else None,
                exception=http_error,
            )
        except RequestException as request_error:
            # Handle other requests-related errors (e.g., network issues)
            error_message = f"Request failed: {request_error}"
            raise DapiError(
                message=error_message,
                status_code=response.status_code if response is not None else None,
                exception=request_error,
            )
        except Exception as e:
            # Catch-all for any other exceptions
            raise DapiError(
                message="An unexpected error occurred", exception=f"e: {e}, traceback: {traceback.format_exc()}"
            )
